/**
 * @(#)Examples.java  10 November 2006
 * Lab 10
 * @author Viera K. Proulx
 */
import java.util.*;
/**
 * Class to represent sample data and test suite for this lab.
 */
public class Examples extends TraversalTestHarness{
  
  /** sample <CODE>{@link Album Album}</CODE>. */
  Album soul1 = new Album("Let's Get It On", "Marvin Gaye", 1973);
  /** sample <CODE>{@link Album Album}</CODE>. */
  Album pop1 = new Album("Purple Rain", "Prince", 1984);
  /** sample <CODE>{@link Album Album}</CODE>. */
  Album rock1 = new Album("Are You Experienced", "Jimi Hendrix", 1966);
  /** sample <CODE>{@link Album Album}</CODE>. */
  Album jazz1 = new Album("Miles Smiles", "Miles Davis", 1966);
  
  /** sample <CODE{@link ArrayList ArrayList}</CODE> of 
   * <CODE>{@link Album Album}</CODE>. */
  ArrayList<Album> albums = new ArrayList<Album>();
  
  /** initialize the CODE{@link ArrayList ArrayList}</CODE> of 
   * <CODE>{@link Album Album}</CODE>. */
  private void makeList(){
    albums.add(soul1);
    albums.add(pop1);
    albums.add(rock1);
    albums.add(jazz1);
  }
  
  /** 
   * Tests for the orMap method and its variants
   */
  private void testOrMaps(){
    makeList();
    
    // expected true:
    test("orMap using ArrayListTraversal",
        true, 
        Algorithms.orMap(new AlbumListTraversal(albums, 0),
            new BeforeYear(1970)));
    
    test("orMap using while loop with Traversal",
        true, 
        Algorithms.orMapWhile(new AlbumListTraversal(albums, 0), 
            new BeforeYear(1970)));
    
    test("orMap using for loop with Traversal",
        true, 
        Algorithms.orMapFor(new AlbumListTraversal(albums, 0), 
            new BeforeYear(1970)));
    
    test("orMap using counted for loop",
        true, 
        Algorithms.orMapForCounted(albums, 
            new BeforeYear(1970)));
    
    // expected false:
    test("orMap using ArrayListTraversal",
        false, 
        Algorithms.orMap(new AlbumListTraversal(albums, 0), 
            new BeforeYear(1950)));
    
    test("orMap using while loop with Traversal",
        false, 
        Algorithms.orMapWhile(new AlbumListTraversal(albums, 0), 
            new BeforeYear(1950)));
    
    test("orMap using for loop with Traversal",
        false, 
        Algorithms.orMapFor(new AlbumListTraversal(albums, 0), 
            new BeforeYear(1950)));
    
    test("orMap using counted for loop",
        false, 
        Algorithms.orMapForCounted(albums, 
            new BeforeYear(1950)));
  }
  
  /**
   * Tests for the mutating methods
   */ 
  private void testMutation(){
    ArrayList<Album> albums = new ArrayList<Album>();
    Album unknown;
    
    //inserts the element at the end of the list
    albums.add(soul1);
    //behaves like cons, inserts the element at the beginning of the list
    albums.add(0, rock1);
    test("contains soul1?", true, albums.contains(soul1));
    test("is the albums list empty?", false, albums.isEmpty());
    
    test("is rock1 the 1st element?", true, rock1.equals(albums.get(0)));
    
    //set changes the object in the given index and 
    //returns the object that was there
    unknown = albums.set(1, pop1);
    test("is pop1 the 2nd element?", true, pop1.equals(albums.get(1)));
    test("before that, was soul1 the 2nd element?", true, soul1.equals(unknown));
    
    albums.add(jazz1);
    test("3 albums in the list?", 3, albums.size());
    
    albums.remove(jazz1);
    test("2 albums in the list?", 2, albums.size());
  }
  
  
  public static void main(String[] args) {
    Examples e = new Examples();
    
    e.testOrMaps();
    e.testMutation();
    
    
    e.fullTestReport();
  }
  
  
}

/**
 * Class that implements <CODE>{@link ISelect ISelect}</CODE> to select
 * <CODE>{@link Album Album}</CODE> that was released in the year 
 * provided to the constructor.
 */
class BeforeYear implements ISelect<Album>{
  int year;
  
  BeforeYear(int year){
    this.year = year;
  }
  
  public boolean select(Album album){
    return album.getYor() < year;
  }
}
